@@ -22,6 +22,8 @@ from pyramid_controllers import \
     includeme, \
     Controller, Dispatcher, \
     expose, index, lookup, default, fiddle, expose_defaults
+from pyramid_controllers.decorator import PCATTR
+from pyramid_controllers.util import getVersion
 from .test_helpers import TestHelper
 
 #------------------------------------------------------------------------------
@@ -32,6 +34,14 @@ class TestDispatcher(TestHelper):
     self._setup(Controller(), '/')
     self.assertEqual([v['route_name'] for v in self.views], ['root-index', 'root'])
 
+  #----------------------------------------------------------------------------
+  def test_version(self):
+    v = getVersion()
+    if v == 'unknown':
+      # todo: this shouldn't really ever happen...
+      return
+    self.assertRegexpMatches(v, '^\d+(\.\d+)*$')
+
   #----------------------------------------------------------------------------
   # TEST @INDEX
   #----------------------------------------------------------------------------
@@ -114,6 +124,21 @@ class TestDispatcher(TestHelper):
         return dict(foo='bar')
     self.assertResponse(self.send(Root(), '/'), 200, "{'foo': 'bar'}")
 
+  def test_index_method(self):
+    '@index can restrict which methods are allowed'
+    class Root(Controller):
+      @index(method='GET')
+      def index_get(self, request):
+        return 'method is GET'
+      @index(method=['put', 'post'])
+      def index_put(self, request):
+        return 'method is %s (PorP)' % (request.method,)
+    self.assertResponse(self.send(Root(), '/'), 200, 'method is GET')
+    self.assertResponse(self.send(Root(), '/', method='PUT'), 200, 'method is PUT (PorP)')
+    self.assertResponse(self.send(Root(), '/', method='POST'), 200, 'method is POST (PorP)')
+    # TODO: this should return a 405...
+    self.assertResponse(self.send(Root(), '/', method='DELETE'), 404)
+
   #----------------------------------------------------------------------------
   # TEST @EXPOSE
   #----------------------------------------------------------------------------
@@ -167,6 +192,37 @@ class TestDispatcher(TestHelper):
         return dict(foo='bar')
     self.assertResponse(self.send(Root(), '/method'), 200, "{'foo': 'bar'}")
 
+  def test_expose_method(self):
+    '@expose can restrict which methods are allowed'
+    class Root(Controller):
+      @expose(name='res', method='GET')
+      def resource_get(self, request):
+        return 'res with GET'
+      @expose(name='res', method=['put', 'post'])
+      def resource_put(self, request):
+        return 'res with %s (PorP)' % (request.method,)
+    self.assertResponse(self.send(Root(), '/res'), 200, 'res with GET')
+    self.assertResponse(self.send(Root(), '/res', method='PUT'), 200, 'res with PUT (PorP)')
+    self.assertResponse(self.send(Root(), '/res', method='POST'), 200, 'res with POST (PorP)')
+    # TODO: this should return a 405...
+    self.assertResponse(self.send(Root(), '/res', method='DELETE'), 404)
+
+  def test_expose_on_bound_method(self):
+    '@expose on a bound method'
+    class Ext(Controller):
+      def __init__(self, name):
+        super(Ext, self).__init__()
+        self.foo  = expose(name=name + '.txt')(self.foo)
+      def foo(self, request): return 'ok.path:' + request.path
+    bar = Ext('bar')
+    zig = Ext('zig')
+    self.assertResponse(self.send(bar, '/foo'),     404)
+    self.assertResponse(self.send(zig, '/foo'),     404)
+    self.assertResponse(self.send(bar, '/bar.txt'), 200, 'ok.path:/bar.txt')
+    self.assertResponse(self.send(zig, '/bar.txt'), 404)
+    self.assertResponse(self.send(bar, '/zig.txt'), 404)
+    self.assertResponse(self.send(zig, '/zig.txt'), 200, 'ok.path:/zig.txt')
+
   #----------------------------------------------------------------------------
   # TEST @EXPOSE ALIASING
   #----------------------------------------------------------------------------
@@ -210,13 +266,6 @@ class TestDispatcher(TestHelper):
     self.assertResponse(self.send(Ext(), '/style.css'),  200, 'COMPILED:ok.name.style')
     self.assertResponse(self.send(Ext(), '/style'),      404)
 
-  def test_expose_name_and_ext_collision(self):
-    '@expose parameters "ext" and "name" are mutually exclusive'
-    with self.assertRaises(ValueError):
-      class Bad(Controller):
-        @expose(name='foo', ext='css')
-        def method(self, request): pass
-
   def test_expose_ext_list(self):
     '@expose parameters "ext" and "name" can accept lists'
     class Ext(Controller):
@@ -230,6 +279,23 @@ class TestDispatcher(TestHelper):
     self.assertResponse(self.send(Ext(), '/bar'),     200, 'ok.name:/bar')
     self.assertResponse(self.send(Ext(), '/name'),    404)
 
+  def test_expose_name_and_ext_are_combinative(self):
+    '@expose parameters "ext" and "name" are combinative'
+    class Ext(Controller):
+      @expose(name=('ver', 'rev'), ext=('txt', 'rst'))
+      def handler(self, request): return 'ok.path:' + request.path
+    # note: this is violating the abstraction barrier... oh well. testing
+    #       the i-rep!... :)
+    self.assertEqual(
+      sorted(getattr(Ext().handler, PCATTR).expose[0]['name']),
+      sorted(['ver.txt', 'ver.rst', 'rev.txt', 'rev.rst']))
+    self.assertResponse(self.send(Ext(), '/ver.txt'), 200, 'ok.path:/ver.txt')
+    self.assertResponse(self.send(Ext(), '/ver.rst'), 200, 'ok.path:/ver.rst')
+    self.assertResponse(self.send(Ext(), '/rev.txt'), 200, 'ok.path:/rev.txt')
+    self.assertResponse(self.send(Ext(), '/rev.rst'), 200, 'ok.path:/rev.rst')
+    self.assertResponse(self.send(Ext(), '/rev'),     404)
+    self.assertResponse(self.send(Ext(), '/ver'),     404)
+
   #----------------------------------------------------------------------------
   # TEST @FIDDLE
   #----------------------------------------------------------------------------
@@ -260,21 +326,51 @@ class TestDispatcher(TestHelper):
       sub = Sub()
     self.assertResponse(self.send(Root(), '/sub/data'), 200, 'ok.sub.data:nsn-fiddled=True')
 
-  # # TODO: this unit test currently fails... re-enable when fixed.
-  # def test_fiddle_with_subclassing(self):
-  #   'Multiple @fiddles coming from both sub-class and super-class should apply'
-  #   class Base(Controller):
-  #     @fiddle
-  #     def basefiddle(self, request):
-  #       request.base = 'y'
-  #   class Sub(Base):
-  #     @fiddle
-  #     def subfiddle(self, request):
-  #       request.sub = 'y'
-  #     @expose
-  #     def chk(self, request):
-  #       return 'ok:base=%s,sub=%s' % (getattr(request, 'base', 'n'), getattr(request, 'sub', 'n'))
-  #   self.assertResponse(self.send(Sub(), '/chk'), 200, 'ok:base=y,sub=y')
+  def test_fiddle_with_subclassing(self):
+    'Multiple @fiddles coming from both sub-class and super-class should apply'
+    class Base(Controller):
+      @fiddle
+      def basefiddle(self, request):
+        request.base = 'y'
+    class Sub(Base):
+      @fiddle
+      def subfiddle(self, request):
+        request.sub = 'y'
+      @expose
+      def chk(self, request):
+        return 'ok:base=%s,sub=%s' % (getattr(request, 'base', 'n'), getattr(request, 'sub', 'n'))
+    self.assertResponse(self.send(Sub(), '/chk'), 200, 'ok:base=y,sub=y')
+
+  def test_fiddle_with_subclassing_and_override(self):
+    'Multiple @fiddles coming from both sub-class and super-class should apply'
+    class Base(Controller):
+      @fiddle
+      def fiddle(self, request):
+        request.base = 'y'
+    class Sub(Base):
+      @fiddle
+      def fiddle(self, request):
+        request.sub = 'y'
+      @expose
+      def chk(self, request):
+        return 'ok:base=%s,sub=%s' % (getattr(request, 'base', 'n'), getattr(request, 'sub', 'n'))
+    self.assertResponse(self.send(Sub(), '/chk'), 200, 'ok:base=n,sub=y')
+
+  def test_fiddle_with_subclassing_and_super_cascade(self):
+    'Multiple @fiddles coming from both sub-class and super-class should apply'
+    class Base(Controller):
+      @fiddle
+      def fiddle(self, request):
+        request.base = 'y'
+    class Sub(Base):
+      @fiddle
+      def fiddle(self, request):
+        super(Sub, self).fiddle(request)
+        request.sub = 'y'
+      @expose
+      def chk(self, request):
+        return 'ok:base=%s,sub=%s' % (getattr(request, 'base', 'n'), getattr(request, 'sub', 'n'))
+    self.assertResponse(self.send(Sub(), '/chk'), 200, 'ok:base=y,sub=y')
 
   #----------------------------------------------------------------------------
   # TEST @LOOKUP
@@ -348,6 +444,21 @@ class TestDispatcher(TestHelper):
         return dict(args=[curpath, list(remainder)])
     self.assertResponse(self.send(Root(), '/zag/zig/zog'), 200, "{'args': ['zag', ['zig', 'zog']]}")
 
+  def test_default_method(self):
+    '@default can restrict which methods are allowed'
+    class Root(Controller):
+      @default(method='GET')
+      def default_get(self, request, path, *rem):
+        return 'default GET'
+      @default(method=['put', 'post'])
+      def default_put(self, request, path, *rem):
+        return 'default %s (PorP)' % (request.method,)
+    self.assertResponse(self.send(Root(), '/res'), 200, 'default GET')
+    self.assertResponse(self.send(Root(), '/res', method='PUT'), 200, 'default PUT (PorP)')
+    self.assertResponse(self.send(Root(), '/res', method='POST'), 200, 'default POST (PorP)')
+    # TODO: this should return a 405...
+    self.assertResponse(self.send(Root(), '/res', method='DELETE'), 404)
+
   #----------------------------------------------------------------------------
   # TEST CUSTOM DISPATCHER
   #----------------------------------------------------------------------------
@@ -404,6 +515,56 @@ class TestDispatcher(TestHelper):
     self.assertResponse(self.send(Root(), '/'),        200, "{'foo': 'idx'}")
     self.assertResponse(self.send(Root(), '/zig-zag'), 200, "{'foo': 'zig-zag'}")
 
+  def test_expose_defaults_renderer_with_subclassing(self):
+    @expose_defaults(renderer='repr')
+    class Base(Controller):
+      @expose
+      def rep(self, request): return dict(m='rep')
+      @expose(renderer='raw')
+      def raw(self, request): return dict(m='raw')
+    class Root(Base):
+      @expose
+      def srep(self, request): return dict(m='srep')
+      @expose(renderer='raw')
+      def sraw(self, request): return dict(m='sraw')
+    def raw(info):
+      def _render(value, system):
+        return 'RAW:' + repr(value)
+      return _render
+    self.renderers['raw'] = raw
+    self.assertResponse(self.send(Root(), "/rep"),  200, "{'m': 'rep'}")
+    self.assertResponse(self.send(Root(), "/raw"),  200, "RAW:{'m': 'raw'}")
+    self.assertResponse(self.send(Root(), "/srep"), 200, "{'m': 'srep'}")
+    self.assertResponse(self.send(Root(), "/sraw"), 200, "RAW:{'m': 'sraw'}")
+
+  # def test_expose_defaults_does_not_pollute(self):
+  #   class Base(Controller):
+  #     def __init__(self):
+  #       super(Base,self).__init__()
+  #       self.count = 0
+  #     @expose
+  #     def m(self, request):
+  #       self.count += 1
+  #       return dict(c=self.count)
+  #   @expose_defaults(renderer='raw')
+  #   class Raw(Base):
+  #     pass
+  #   @expose_defaults(renderer='repr')
+  #   class Rep(Base):
+  #     pass
+  #   class Root(Controller):
+  #     raw = Raw()
+  #     rep = Rep()
+  #   def raw(info):
+  #     def _render(value, system):
+  #       return 'RAW:' + repr(value)
+  #     return _render
+  #   self.renderers['raw'] = raw
+  #   root = Root()
+  #   self.assertResponse(self.send(root, '/raw/m'), 200, "RAW:{'c': 1}")
+  #   self.assertResponse(self.send(root, '/rep/m'), 200, "{'c': 2}")
+  #   self.assertResponse(self.send(root, '/raw/m'), 200, "RAW:{'c': 3}")
+
   def test_expose_defaults_ext(self):
     'Controllers can specify default extensions for member @expose calls'
     @expose_defaults(ext='json')
@@ -414,7 +575,7 @@ class TestDispatcher(TestHelper):
       def bar(self, request): return 'bar'
       @expose(ext='js')
       def zag(self, request): return 'zag'
-      @expose(name='zig.css')
+      @expose(name='zig.css', ext=None)
       def zig(self, request): return 'zig'
       @expose(ext=None)
       def zog(self, request): return 'zog'
@@ -433,6 +594,77 @@ class TestDispatcher(TestHelper):
     self.assertResponse(self.send(Root(), '/zog.js'),   404)
     self.assertResponse(self.send(Root(), '/zog.css'),  404)
 
+  def test_expose_defaults_with_subclassing_from_base(self):
+    @expose_defaults(ext='json')
+    class Base(Controller):
+      @expose
+      def foo(self, request): return 'foo'
+    class Root(Base):
+      @expose
+      def bar(self, request): return 'bar'
+    self.assertResponse(self.send(Root(), '/foo.json'), 200, 'foo')
+    self.assertResponse(self.send(Root(), '/foo'),      404)
+    self.assertResponse(self.send(Root(), '/bar.json'), 200, 'bar')
+    self.assertResponse(self.send(Root(), '/bar'),      404)
+
+  def test_expose_defaults_with_subclassing_to_base(self):
+    class Base(Controller):
+      @expose
+      def foo(self, request): return 'foo'
+    @expose_defaults(ext='json')
+    class Root(Base):
+      @expose
+      def bar(self, request): return 'bar'
+    self.assertResponse(self.send(Root(), '/foo.json'), 200, 'foo')
+    self.assertResponse(self.send(Root(), '/foo'),      404)
+    self.assertResponse(self.send(Root(), '/bar.json'), 200, 'bar')
+    self.assertResponse(self.send(Root(), '/bar'),      404)
+
+  def test_expose_defaults_with_subclassing_and_override(self):
+    @expose_defaults(ext='json')
+    class Base(Controller):
+      @expose
+      def foo(self, request): return 'foo'
+    @expose_defaults(ext='js')
+    class Root(Base):
+      @expose
+      def bar(self, request): return 'bar'
+    self.assertResponse(self.send(Root(), '/foo.json'), 200, 'foo')
+    self.assertResponse(self.send(Root(), '/foo'),      404)
+    self.assertResponse(self.send(Root(), '/bar.js'),   200, 'bar')
+    self.assertResponse(self.send(Root(), '/bar.json'), 404)
+    self.assertResponse(self.send(Root(), '/bar'),      404)
+
+  def test_expose_defaults_with_override(self):
+    @expose_defaults(ext='json')
+    class Root(Controller):
+      @expose
+      def foo(self, request): return 'foo'
+      @expose(ext=None)
+      def bar(self, request): return 'bar'
+      @expose(ext=(None, 'txt'))
+      def zig(self, request): return 'zig:' + request.path
+    self.assertResponse(self.send(Root(), '/foo.json'), 200, 'foo')
+    self.assertResponse(self.send(Root(), '/foo'),      404)
+    self.assertResponse(self.send(Root(), '/bar'),      200, 'bar')
+    self.assertResponse(self.send(Root(), '/bar.json'), 404)
+    self.assertResponse(self.send(Root(), '/zig'),      200, 'zig:/zig')
+    self.assertResponse(self.send(Root(), '/zig.txt'),  200, 'zig:/zig.txt')
+    self.assertResponse(self.send(Root(), '/zig.json'), 404)
+
+  def test_expose_defaults_with_multiple_names(self):
+    @expose_defaults(ext=('json', 'yaml'))
+    class Root(Controller):
+      @expose(name=('blue', 'moon'))
+      def foo(self, request): return 'path:' + request.path
+    self.assertEqual(
+      sorted(getattr(Root().foo, PCATTR).expose[0]['name']),
+      sorted(['blue.json', 'blue.yaml', 'moon.json', 'moon.yaml']))
+    self.assertResponse(self.send(Root(), '/blue.html'), 404)
+    self.assertResponse(self.send(Root(), '/blue'),      404)
+    self.assertResponse(self.send(Root(), '/blue.json'), 200, 'path:/blue.json')
+    self.assertResponse(self.send(Root(), '/blue.yaml'), 200, 'path:/blue.yaml')
+
 #------------------------------------------------------------------------------
 # end of $Id$
 #------------------------------------------------------------------------------
